
package kdocCxxDocbook;
use Carp;
use File::Path;

use strict;
no strict "subs";

use vars qw/ $lib $rootnode $outputdir $opt @docqueue /;

# public, protected, private
# within each:
	# types/typedefs
	# compounds
	# properties
	# methods
	# statics
	
sub writeDoc
{
	( $lib, $rootnode, $outputdir, $opt ) = @_;

	@docqueue = ();

	mkpath( $outputdir) unless -f $outputdir;

	open ( DOC, ">$outputdir/$lib.sgml" ) || die "Couldn't write output.";

	my $time = localtime; chop $time;

print DOC<<EOF;
<!doctype book PUBLIC "-//Davenport//DTD DocBook V3.0//EN" [
]>
<book id="$lib-lib">
	<bookinfo>
		<date>$time</date>
		<title>$lib API Documentation</title>
	</bookinfo>
	<toc></toc>
EOF

	printHierarchy();
	printClassIndex();
	printClassDoc();

print DOC<<EOF;
</book>
EOF

}



sub printHierarchy
{

	print DOC chapter( "class-hierarchy", "$lib Class Hierarchy" );
	printNodeHier( $rootnode );

	print DOC tclose( "chapter" ); 
}

sub printNodeHier
{
	my ( $node ) = @_;
	my $src = "";

	if ( defined $node->{ExtSource} ) {
		return if !defined $node->{InBy}
		|| !kdocAstUtil::hasLocalInheritor( $node );

		$src = "(".$node->{ExtSource}.")";
	}

	print DOC "<ListItem><para>", refName($node)," $src</para>\n"
		unless $node == $rootnode;

	if ( defined $node->{InBy} ) {

		print DOC "<ItemizedList>\n";

		foreach my $kid ( 
				sort { $a->{astNodeName} cmp $b->{astNodeName} }
				@{ $node->{InBy} } ) {
			next if defined $kid->{hlisted};
			printNodeHier( $kid );
			$kid->AddProp( 'hlisted', 1 );
		}
		print DOC "</ItemizedList>\n";

	}
	print DOC "</ListItem>\n" unless $node == $rootnode;	
}

sub printClassIndex
{
	print DOC chapter( "class-index", "$lib Class Index" );

	
	print DOC topen( "table", "TocEntry", 1, "PgWide", 1, "ColSep", 0 ),
		tblk( "title", $lib.' classes' ),
		tblk( "tgroup" );

	foreach my $kid ( sort { $a->{astNodeName} cmp $b->{astNodeName} }
			@{ $rootnode->{Kids} } ) {
		next unless kdocAstUtil::localComp( $kid );

		print DOC '<ROW><ENTRY>', refName( $kid ),'</ENTRY>',
			'<ENTRY>', deref( $kid->{DocNode}->{ClassShort},
				$rootnode), 
				tclose( "Entry", "Row" );
	}


	print DOC tclose( "TBody", "TGroup", "Table", "chapter" );
}

sub printClassDoc
{
	print DOC chapter( "class-doc", "$lib Class Documentation" );

	push @docqueue, $rootnode;

	while( $#docqueue >= 0 ) {
		my $node = pop @docqueue;
		docChildren( $node );
	}

	print DOC "</chapter>\n";
}

sub docChildren
{
	my ( $node ) = @_;

	## Queue local compound children
	
	return if !defined $node->{Kids};

	foreach my $kid ( @{$node->{Kids}} ) {
		next unless kdocAstUtil::localComp( $kid );
		
		push @docqueue, $kid;
	}

	return if $node == $rootnode || !kdocAstUtil::hasDoc( $node );
	
	
	## document self and children
	
	print DOC '<Sect1 id="doc-', fullName( $node, '-' ), 
		'" XRefLabel="', fullName( $node, "::") ,'">', "\n",
		tblk( "title", $node->{NodeType}." ".fullName( $node ) );

	# TODO Warnings, notes etc

	printDoc ( $node->{DocNode}, *DOC, $rootnode, 1 );

	# TODO: Member docs go here.

	# First a member index by type
	# Then long docs for each member
	
	print DOC "</Sect1>\n";

}

=head2 printDoc

	Parameters: docnode, *filehandle, rootnode, compound

	Print a doc node. If compound is specified and non-zero, various
	compound node properties are not printed.

=cut

sub printDoc
{
	my $docNode = shift;
	local *CLASS = shift;

        my ( $rootnode, $comp ) = @_;
        my $node;
        my $type;
	my $text;
	my $lasttype = "none";

	$comp = 0 if !defined $comp;
	
	$text = $docNode->{Text};

        if ( defined $text ) {
                print CLASS "<para>";

                foreach $node ( @$text ) {
                        $type = $node->{NodeType};
			my $name = $node->{astNodeName};
                        warn "Node '", $name, "' has no type"
				if !defined $type;

			if( $lasttype eq "ListItem" && $type ne $lasttype ) {
				print CLASS "</ItemizedList><para>\n";
			}

                        if( $type eq "DocText" ) {
                                print CLASS "", deref( $name, $rootnode );
                        }
                        elsif ( $type eq "Pre" ) {
                                print CLASS "</para>",
					tblk( "ProgramListing", esc( $name ) ), 
						"<para>";
                        }
			elsif( $type eq "Ref" ) {
				my $ref = $node->{Ref};
				if ( defined $ref ) {
					print "found reference for $name\n";
					print CLASS refName( $ref );
				}
				else {
					print CLASS $name;
				}
			}
			elsif ( $type eq "Image" ) {
				print CLASS "<Graphic FileRef=\"",
					$node->{Path}, "\"></Graphic>";
			}
                        elsif ( $type eq "ParaBreak" ) {
                                print CLASS "</para><para>";
                        }
			elsif ( $type eq "ListItem" ) {
				if ( $lasttype ne "ListItem" ) {
					print CLASS "</para><ItemizedList>\n";
				}
				print CLASS tblk( "ListItem",
					tblk("para",deref($name,$rootnode )) );
			}

			$lasttype = $type;
                }

		if( $type eq "ListItem" ) {
			print CLASS "</ListItem><para>\n";
		}

                print CLASS "</para>";

        }

	# Params
	my @paramlist = ();
	kdocAstUtil::findNodes( \@paramlist, $docNode->{Text},
			"NodeType", "Param" );

	if( $#paramlist >= 0 ) {
		my $pnode;
		print CLASS "<p><b>Parameters</b>:",
			"<TABLE BORDER=\"0\" CELLPADDING=\"5\">\n";

		print CLASS topen( 'Table', 'PgWide', "1", "colsep", "0" ),
			tblk( 'Title', $lib. ' classes' ),
			topen( "TGroup", "cols", "2" ), tclose( "TBody>" );

		foreach $pnode ( @paramlist ) {
			print CLASS "<ROW><ENTRY>", esc($pnode->{Name}),
				"</ENTRY><ENTRY>",
				deref($pnode->{astNodeName}, $rootnode ),
				"</ROW>\n";
		}
		print CLASS tclose( "TBody", "TGroup", "Table" );
	}

	# Return
	printTextItem( $docNode, CLASS, "Returns" );

	# Exceptions
	$text = $docNode->{Throws};

	if ( defined $text ) {
		my $comma = "<formalpara><title>Exceptions</title><para>";

		foreach my $tosee ( @$text ) {
			print CLASS $comma, esc( $tosee );
			$comma = ", ";
		}
		print DOC tclose( "para", "formalpara" );
	}

	# See
	$text = $docNode->{See};
	my $tref = $docNode->{SeeRef};
	
	if ( defined $text ) {
		my $comma = "<tip><title>See also..</title><para>";
		my $ctr = -1;

		foreach $ctr ( 0..$#{$text} ) {
			if ( defined $tref->[ $ctr ] ) {
				print CLASS $comma, refName( $tref->[ $ctr ] );
			}
			else {
				print CLASS $comma, esc( $text->[ $ctr ] );
			}

			$comma = ", ";
		}
		print CLASS tclose( "para", "tip" ) unless $ctr < 0;
	}

	return if $comp;

	printTextItem( $docNode, CLASS, "Since" );
	printTextItem( $docNode, CLASS, "Version" );
	printTextItem( $docNode, CLASS, "Id" );
	printTextItem( $docNode, CLASS, "Author" );
}

sub printTextItem
{
	my $node = shift;
	local *CLASS = shift;
	my ( $prop, $label ) = @_;

	my $text = $node->{ $prop };
	
	return unless defined $text;
	$label = $prop unless defined $label;
	
	print CLASS "<formalpara><title>", $label, "</title><para> ", 
		deref( $text, $rootnode  ), "</para></formalpara>\n";
}


# utilities

sub chapter
{
	my ( $id, $title ) = @_;

	return "<chapter id=\"$id\"><title>$title</title>";
}


sub refName
{
	my( $node ) = @_;

	return fullName( $node ) if defined $node->{ExtSource}
		|| !kdocAstUtil::hasDoc( $node );

	return "<xref LinkEnd=\"doc-".fullName( $node, "--")."\">";
}

sub fullName
{
	my( $node, $sep ) = @_;

	return $node->{astNodeName} if $node == $rootnode 
		|| $node->{Parent} == $rootnode;
	return fullName( $node->{Parent}, $sep ).$sep.$node->{astNodeName};
}

sub print
{
	
}
=head2 deref

	Parameters: text, rootnode
	returns text

	dereferences all @refs in the text and returns it.

=cut

sub deref
{
	my ( $str, $rootnode ) = @_;
	confess "rootnode is null" if !defined $rootnode;
	my $out = "";
	my $text;

	foreach $text ( split (/(\@\w+\s+[\w:#]+)/, $str ) ) {
		if ( $text =~ /\@ref\s+([\w:#]+)/ ) {
			my $name = $1;
			$name =~ s/^\s*#//g;
			$out .= wordRef( $name, $rootnode );
		}
		elsif ( $text =~ /\@p\s+([\w:]+)/ ) {
			$out .= tblk( "Literal", esc($1) );
		}
		else {
			$out .= esc($text);
		}
	}

	return $out;
}



=head3 wordRef

	Parameters: word

	Prints a hyperlink to the word's reference if found, otherwise
	just prints the word. Good for @refs etc.

=cut

sub wordRef
{
	my ( $str, $rootnode ) = @_;
	confess "rootnode is undef" if !defined $rootnode;

	return "" if $str eq "";

	my $ref = kdocAstUtil::findRef( $rootnode, $str );
	return esc($str) if !defined $ref;

	return refName( $ref );
}


=head2 esc

	Escape special SGML characters for normal text.

=cut

sub esc
{
	my $str = $_[ 0 ];

	return "" if !defined $str || $str eq "";

	$str =~ s/&/&amp;/g;
	$str =~ s/</&lt;/g;
	$str =~ s/>/&gt;/g;

	return $str;
}


sub tblk
{
	my ( $tag, $text );
	return "<$tag>".$text."</$tag>";
}


sub topen
{
	my $tag = shift;

	carp "mismatched ids and tags" if ($#_+1) % 2 != 0;

	my $out = "<$tag";

	while ( $#_ >= 0 ) {
		$out .= " ".shift( @_ ).'="'.shift( @_ ).'"';
	}

	$out .= ">";

	return $out;
}

sub tclose
{
	my $out = "";

	foreach my $tag ( @_ ) {
		$out .= "</$tag>";
	}

	return $out;
}

1;
